@@ -3,6 +3,8 @@
 
 from itertools import islice
 
+UNDEFINED = object()
+
 
 def aggregate(sequence, func, seed=None):
     """
@@ -34,6 +36,84 @@ def aggregate(sequence, func, seed=None):
     return res
 
 
+def argmax(sequence, key=None):
+    """
+    Returns index of the maximum item in a sequence by using default comparer
+    or specified item's key.
+    
+    Args:
+        sequence: iterable
+            Sequence of items to sort.
+        
+        key: callable or None
+            Item's key selector.
+    
+    Returns:
+        int
+            Index of the maximum item.
+    """
+    
+    items = [key(d) for d in sequence] if key is not None else sequence
+    
+    if not hasattr(items, '__len__'):
+        items = list(items)
+    
+    return max(range(len(items)), key=items.__getitem__)
+
+
+def argmin(sequence, key=None):
+    """
+    Returns index of the minimum item in a sequence by using default comparer
+    or specified item's key.
+    
+    Args:
+        sequence: iterable
+            Sequence of items to sort.
+        
+        key: callable or None
+            Item's key selector.
+    
+    Returns:
+        int
+            Index of the minimum item.
+    """
+    
+    items = [key(d) for d in sequence] if key is not None else sequence
+    
+    if not hasattr(items, '__len__'):
+        items = list(items)
+    
+    return min(range(len(items)), key=items.__getitem__)
+
+
+def argsort(sequence, key=None, reverse=False):
+    """
+    Returns items indices that would sort a sequence by using default comparer
+    or specified item's key.
+    
+    Args:
+        sequence: iterable
+            Sequence of items to sort.
+        
+        key: callable or None
+            Item's key selector.
+        
+        reverse: bool
+            If set to True, sorting is reversed.
+    
+    Returns:
+        (int,)
+            Indexes of sorted items.
+    """
+    
+    items = [key(d) for d in sequence] if key is not None else sequence
+    
+    if not hasattr(items, '__len__'):
+        items = list(items)
+    
+    return sorted(range(len(items)), key=items.__getitem__, reverse=reverse)
+
+
 def bisect(sequence, value, key=None, side='left'):
     """
     Uses binary search to find index where if given value inserted, the order
@@ -147,38 +227,16 @@ def count(sequence, condition=None):
     return sum(1 for d in sequence if condition(d))
 
 
-def distinct(sequence):
+def distinct(sequence, key=None):
     """
-    Iterates over distinct items in a sequence by using default comparer. First
-    occurrence of each item is used.
-    
-    Args:
-        sequence: iterable
-            Sequence of items to go through.
-    
-    Returns:
-        iter(any)
-            Iterator over distinct items.
-    """
-    
-    seen = set()
-    
-    for item in sequence:
-        if item not in seen:
-            seen.add(item)
-            yield item
-
-
-def distinct_by(sequence, key):
-    """
-    Iterates over distinct items in a sequence by using specified item's key.
-    First occurrence of each item is used.
+    Iterates over distinct items in a sequence by using default comparer or
+    specified item's key. First occurrence of each item is used.
     
     Args:
         sequence: iterable
             Sequence of items to go through.
         
-        key: callable
+        key: callable or None
             Item's key selector.
     
     Returns:
@@ -186,41 +244,21 @@ def distinct_by(sequence, key):
             Iterator over distinct items.
     """
     
+    has_key = key is not None
     seen = set()
     
     for item in sequence:
-        k = key(item)
+        k = key(item) if has_key else item
+        
         if k not in seen:
             seen.add(k)
             yield item
 
 
-def exclude(sequence, items):
+def exclude(sequence, items, key=None):
     """
-    Excludes specified items from a sequence by using default comparer.
-    
-    Args:
-        sequence: iterable
-            Sequence of items to go through.
-        
-        items: iterable
-            Items to exclude.
-    
-    Returns:
-        iter(any)
-            Iterator over remaining items.
-    """
-    
-    keys = set((d for d in items))
-    
-    for item in sequence:
-        if item not in keys:
-            yield item
-
-
-def exclude_by(sequence, items, key):
-    """
-    Excludes specified items from a sequence by using selected item's key.
+    Excludes specified items from a sequence by using default comparer or
+    specified item's key.
     
     Args:
         sequence: iterable
@@ -229,7 +267,7 @@ def exclude_by(sequence, items, key):
         items: iterable
             Items to exclude.
         
-        key: callable
+        key: callable or None
             Item's key selector.
     
     Returns:
@@ -237,41 +275,20 @@ def exclude_by(sequence, items, key):
             Iterator over remaining items.
     """
     
-    keys = set((key(d) for d in items))
+    has_key = key is not None
+    keys = set((key(d) if has_key else d for d in items))
     
     for item in sequence:
-        k = key(item)
+        k = key(item) if has_key else item
+        
         if k not in keys:
             yield item
 
 
-def first(sequence, condition=None):
+def first(sequence, condition=None, default=UNDEFINED):
     """
     Returns the first item in a sequence that satisfies specified condition or
-    raises error if no item found.
-    
-    Args:
-        sequence: iterable
-            Sequence of items to go through.
-        
-        condition: callable
-            Condition to test.
-    
-    Returns:
-        any
-            First valid item.
-    """
-    
-    if condition is None:
-        return next((d for d in sequence))
-    
-    return next((d for d in sequence if condition(d)))
-
-
-def first_or_default(sequence, condition=None, default=None):
-    """
-    Returns the first item in a sequence that satisfies specified condition or
-    specified default value if no item found.
+    raises error if no item found and no default value is provided.
     
     Args:
         sequence: iterable
@@ -288,49 +305,27 @@ def first_or_default(sequence, condition=None, default=None):
             First valid item.
     """
     
-    if condition is None:
-        return next((d for d in sequence), default)
+    if condition is not None:
+        items = (d for d in sequence if condition(d))
+    else:
+        items = (d for d in sequence)
     
-    return next((d for d in sequence if condition(d)), default)
+    if default is not UNDEFINED:
+        return next(items, default)
+    
+    return next(items)
 
 
-def group(sequence):
+def group(sequence, key=None):
     """
-    Groups items of a sequence according to default comparer and creates result
-    values as (key, group) pairs.
-    
-    Args:
-        sequence: iterable
-            Sequence of items to go through.
-    
-    Returns:
-        ((any, (any,)),)
-            Grouped items as (key, group) pairs.
-    """
-    
-    keys = []
-    groups = {}
-    
-    for item in sequence:
-        if item not in groups:
-            keys.append(item)
-            groups[item] = []
-        
-        groups[item].append(item)
-    
-    return [(k, tuple(groups[k])) for k in keys]
-
-
-def group_by(sequence, key):
-    """
-    Groups items of a sequence according to specified key selector and creates
-    result values as (key, group) pairs.
+    Groups items of a sequence according to default comparer or specified
+    item's key and creates result values as (key, group) pairs.
     
     Args:
         sequence: iterable
             Sequence of items to go through.
         
-        key: callable
+        key: callable or None
             Item's key selector.
     
     Returns:
@@ -338,11 +333,12 @@ def group_by(sequence, key):
             Grouped items as (key, group) pairs.
     """
     
+    has_key = key is not None
     keys = []
     groups = {}
     
     for item in sequence:
-        k = key(item)
+        k = key(item) if has_key else item
         
         if k not in groups:
             keys.append(k)
@@ -370,42 +366,19 @@ def index(sequence, condition):
             Index of the first valid item.
     """
     
-    for i, item in enumerate(sequence):
+    i = -1
+    for item in sequence:
+        i += 1
         if condition(item):
             return i
     
     raise ValueError()
 
 
-def intersect(sequence, items):
+def intersect(sequence, items, key=None):
     """
     Produces a sequence of shared unique items from given sequences by using
-    default comparer.
-    
-    Args:
-        sequence: iterable
-            Sequence of items to go through.
-        
-        items: iterable
-            Items to intersect with.
-    
-    Returns:
-        iter(any)
-            Iterator over intersecting items.
-    """
-    
-    keys = set((d for d in items))
-    
-    for item in sequence:
-        if item in keys:
-            keys.remove(item)
-            yield item
-
-
-def intersect_by(sequence, items, key):
-    """
-    Produces a sequence of shared unique items from given sequences by using
-    selected item's key.
+    default comparer or specified item's key.
     
     Args:
         sequence: iterable
@@ -414,7 +387,7 @@ def intersect_by(sequence, items, key):
         items: iterable
             Items to intersect with.
         
-        key: callable
+        key: callable or None
             Item's key selector.
     
     Returns:
@@ -422,41 +395,21 @@ def intersect_by(sequence, items, key):
             Iterator over intersecting items.
     """
     
-    keys = set((key(d) for d in items))
+    has_key = key is not None
+    keys = set((key(d) if has_key else d for d in items))
     
     for item in sequence:
-        k = key(item)
+        k = key(item) if has_key else item
+        
         if k in keys:
             keys.remove(k)
             yield item
 
 
-def last(sequence, condition=None):
+def last(sequence, condition=None, default=UNDEFINED):
     """
     Returns the last item in a sequence that satisfies specified condition or
-    raises error if no item found.
-    
-    Args:
-        sequence: iterable
-            Sequence of items to go through.
-        
-        condition: callable
-            Condition to test.
-    
-    Returns:
-        any
-            Last valid item.
-    """
-    
-    sequence = reversed(list(sequence))
-    
-    return first(sequence, condition)
-
-
-def last_or_default(sequence, condition=None, default=None):
-    """
-    Returns the last item in a sequence that satisfies specified condition or
-    specified default value if no item found.
+    raises error if no item found and no default value is provided.
     
     Args:
         sequence: iterable
@@ -473,15 +426,172 @@ def last_or_default(sequence, condition=None, default=None):
             Last valid item.
     """
     
-    sequence = reversed(list(sequence))
+    if condition is not None:
+        items = (d for d in sequence if condition(d))
+    else:
+        items = (d for d in sequence)
     
-    return first_or_default(sequence, condition, default)
+    item = default
+    for item in items:
+        pass
+    
+    if item is UNDEFINED:
+        raise StopIteration
+    
+    return item
 
 
-def single(sequence, condition=None):
+def multisort(sequence, key=None, reverse=False, _n=0):
+    """
+    Produces new sequence by sorting elements of current sequence by using
+    default comparer or specified item's key. If the key provides multiple
+    columns, the sorting direction can be specified for each individual column.
+    
+    Args:
+        sequence: iterable
+            Sequence of items to sort.
+        
+        key: callable or None
+            Item's key selector.
+        
+        reverse: bool or (bool,)
+            If set to True, sorting is reversed. This flag can be specified
+            independently foreach key column.
+    
+    Returns:
+        list(any)
+            Sorted sequence.
+    """
+    
+    # simple sort
+    if reverse is True or reverse is False:
+        return sorted(sequence, key=key, reverse=reverse)
+    
+    # initial sort
+    sequence = sorted(sequence, key=key, reverse=reverse[_n])
+    
+    # no need for further sorts
+    if len(sequence) < 2 or all(reverse[_n:]) or not any(reverse[_n:]):
+        return sequence
+    
+    # get keys
+    keys = sequence if key is None else list(map(key, sequence))
+    
+    # re-sort same keys
+    final = []
+    i = 0
+    while i < len(keys):
+        
+        # get same
+        k = keys[i][:_n+1]
+        j = i+1
+        while j < len(keys) and k == keys[j][:_n+1]:
+            j += 1
+        
+        # sort same
+        if j - i > 1:
+            final += multisort(sequence[i:j], key, reverse, _n+1)
+        else:
+            final.append(sequence[i])
+        
+        i = j
+    
+    return final
+
+
+def rank(sequence, key=None, method='average', reverse=False):
+    """
+    Provides 1-based rank for each item of a sequence by using default
+    comparer or specified item's key. The ties are resolved according to
+    selected method.
+    
+    Args:
+        sequence: iterable
+            Sequence of items to go through.
+        
+        key: callable or None
+            Item's key selector.
+        
+        method: str
+            Method used to assign ranks to tied items.
+                'average' - tied values are assigned by their average rank
+                'min' - tied values are assigned by their minimum rank
+                'max' - tied values are assigned by their maximum rank
+                'dense' - like 'min' but without rank gaps
+                'ordinal' - distinct rank for all values
+        
+        reverse: bool
+            If set to True, sorting is reversed.
+    
+    Returns:
+        (int,)
+            Items' ranks.
+    """
+    
+    items = [key(d) for d in sequence] if key is not None else sequence
+    
+    if not hasattr(items, '__len__'):
+        items = list(items)
+    
+    size = len(items)
+    idxs = sorted(range(size), key=items.__getitem__, reverse=reverse)
+    
+    items = [items[r] for r in idxs]
+    ranks = [0]*size
+    
+    dupls = 0
+    sums = 0
+    lost = 0
+    
+    if method == 'average':
+        for i in range(size):
+            dupls += 1
+            sums += i
+            if i == size - 1 or items[i] != items[i + 1]:
+                for j in range(i - dupls + 1, i + 1):
+                    ranks[idxs[j]] = sums / dupls + 1
+                dupls = 0
+                sums = 0
+    
+    elif method == 'min':
+        for i in range(size):
+            dupls += 1
+            if i == size - 1 or items[i] != items[i + 1]:
+                for j in range(i - dupls + 1, i + 1):
+                    ranks[idxs[j]] = i + 1 - dupls + 1
+                dupls = 0
+    
+    elif method == 'max':
+        for i in range(size):
+            dupls += 1
+            if i == size - 1 or items[i] != items[i + 1]:
+                for j in range(i - dupls + 1, i + 1):
+                    ranks[idxs[j]] = i + 1
+                dupls = 0
+    
+    elif method == 'dense':
+        for i in range(size):
+            dupls += 1
+            sums += i
+            if i == size - 1 or items[i] != items[i + 1]:
+                for j in range(i - dupls + 1, i + 1):
+                    ranks[idxs[j]] = i + 1 - lost
+                dupls = 0
+            else:
+                lost += 1
+    
+    elif method == 'ordinal':
+        for i in range(size):
+            ranks[idxs[i]] = i + 1
+    
+    return ranks
+
+
+def single(sequence, condition=None, default=UNDEFINED):
     """
     Returns the single item in a sequence that satisfies specified condition or
-    raises error if none or more items found.
+    raises error if more items found. If no item is found, returns default
+    value if provided or raises error.
     
     Args:
         sequence: iterable
@@ -489,16 +599,26 @@ def single(sequence, condition=None):
         
         condition: callable
             Condition to test.
+        
+        default: any
+            Default value.
     
     Returns:
         any
             Single valid item.
     """
     
-    items = list(sequence) if condition is None else [d for d in sequence if condition(d)]
+    items = [d for d in sequence if condition(d)] if condition is not None else sequence
+    
+    if not hasattr(items, '__len__'):
+        items = list(items)
+    
     if len(items) == 1:
         return items[0]
     
+    if len(items) == 0 and default is not UNDEFINED:
+        return default
+    
     raise ValueError
 
 
@@ -597,35 +717,10 @@ def take_while(sequence, condition):
             return
 
 
-def union(sequence, items):
+def union(sequence, items, key=None):
     """
     Produces a sequence of unique items from given sequences by using default
-    comparer.
-    
-    Args:
-        sequence: iterable
-            Sequence of items to go through.
-        
-        items: iterable
-            Items to union with.
-    
-    Returns:
-        iter(any)
-            Iterator over unionized items.
-    """
-    
-    keys = set()
-    
-    for item in concat(sequence, items):
-        if item not in keys:
-            keys.add(item)
-            yield item
-
-
-def union_by(sequence, items, key):
-    """
-    Produces a sequence of unique items from given sequences by using selected
-    item's key.
+    comparer or specified item's key.
     
     Args:
         sequence: iterable
@@ -634,7 +729,7 @@ def union_by(sequence, items, key):
         items: iterable
             Items to union with.
         
-        key: callable
+        key: callable or None
             Item's key selector.
     
     Returns:
@@ -642,10 +737,12 @@ def union_by(sequence, items, key):
             Iterator over unionized items.
     """
     
+    has_key = key is not None
     keys = set()
     
     for item in concat(sequence, items):
-        k = key(item)
+        k = key(item) if has_key else item
+        
         if k not in keys:
             keys.add(k)
             yield item
