@@ -2,6 +2,7 @@
 #  Copyright (c) Martin Strohalm. All rights reserved.
 
 import statistics
+import random
 from . import iters
 
 
@@ -9,7 +10,7 @@ class Linque(object):
     """
     Linque provides a functionality similar to .NET LINQ (Language Integrated
     Query) and IEnumerable. It allows chaining multiple queries on a sequence of
-    items, mostly without evaluating the sequence until necessary. Therefore
+    items, mostly without evaluating the sequence until necessary. Therefore,
     each linq chain can only be called once to get a final result. To break this
     logic, the class can be initialized with 'evaluate' parameter set to True,
     to make sure all input data are stored as fully evaluated sequence such as
@@ -17,33 +18,32 @@ class Linque(object):
     steps in memory as a list.
     """
     
-    def __init__(self, items, evaluate=False):
+    def __init__(self, source, evaluate=False):
         """
         Initializes a new instance of Linque.
         
         Args:
-            items: iterable
+            source: iterable
                 Sequence of items.
             
             evaluate: bool
-                If set to True, items sequence is evaluated into tuple upon
+                If set to True, items sequence is evaluated into list upon
                 class init and for each method call as well.
         """
         
-        self._items = list(items) if evaluate else items
+        self._source = list(source) if evaluate else source
         self._evaluate = evaluate
     
     
     def __iter__(self):
         """Gets items iterator."""
         
-        return iter(self._items)
+        return iter(self._source)
     
     
     def aggregate(self, accumulator, seed=None):
         """
-        Applies accumulator function over current sequence. This call fully
-        evaluates current sequence.
+        Applies accumulator function over current sequence.
         
         Args:
             accumulator: callable
@@ -63,7 +63,7 @@ class Linque(object):
     def all(self, condition):
         """
         Determines whether all items of current sequence satisfy given
-        condition. This call does not evaluate current sequence.
+        condition.
         
         Args:
             condition: callable
@@ -79,8 +79,7 @@ class Linque(object):
     def any(self, condition=None):
         """
         Determines whether current sequence contains any item or whether any
-        item of current sequence satisfies given condition. This call does not
-        evaluate current sequence.
+        item of current sequence satisfies given condition.
         
         Args:
             condition: callable or None
@@ -96,10 +95,109 @@ class Linque(object):
         return any(condition(d) for d in self)
     
     
+    def argmax(self, key=None):
+        """
+        Returns index of the maximum item in a sequence by using default
+        comparer or specified item's key.
+        
+        Args:
+            key: callable or None
+                Item's key selector.
+        
+        Returns:
+            int
+        """
+        
+        return iters.argmax(self, key)
+    
+    
+    def argmin(self, key=None):
+        """
+        Returns index of the minimum item in a sequence by using default
+        comparer or specified item's key.
+        
+        Args:
+            key: callable or None
+                Item's key selector.
+        
+        Returns:
+            int
+        """
+        
+        return iters.argmin(self, key)
+    
+    
+    def argsort(self, key=None, reverse=False):
+        """
+        Returns items indices that would sort current sequence by using
+        default comparer or specified item's key.
+        
+        Args:
+            key: callable or None
+                Item's key selector.
+            
+            reverse: bool
+                If set to True, sorting is reversed.
+        
+        Returns:
+            Linque
+        """
+        
+        def source():
+            for item in iters.argsort(self, key, reverse=reverse):
+                yield item
+        
+        result = (d for d in source())
+        
+        return Linque(result, self._evaluate)
+    
+    
+    def choice(self, weights=None):
+        """
+        Returns random item from current sequence.
+        
+        Args:
+            weights: (float,) or None
+                Relative probabilities for individual items to be selected.
+        
+        Returns:
+            any
+        """
+        
+        if weights is None:
+            return random.choice(list(self))
+        
+        return random.choices(list(self), weights=weights, k=1)[0]
+    
+    
+    def choices(self, count, weights=None):
+        """
+        Produces new sequence by randomly choosing number of items from current
+        sequence. Each item can be selected multiple times.
+        
+        Args:
+            count: int
+                Number of choices to make.
+            
+            weights: (float,) or None
+                Relative probabilities for individual items to be selected.
+        
+        Returns:
+            Linque
+        """
+        
+        def source():
+            for item in random.choices(list(self), weights=weights, k=count):
+                yield item
+        
+        result = (d for d in source())
+        
+        return Linque(result, self._evaluate)
+    
+    
     def chunk(self, size):
         """
-        Splits current sequence into chunks of specified size. This call does
-        not evaluate current sequence.
+        Splits current sequence into chunks of specified size.
         
         Args:
             
@@ -119,7 +217,7 @@ class Linque(object):
     def concat(self, items):
         """
         Produces new sequence by appending given items at the end of current
-        sequence. This call does not evaluate current sequence.
+        sequence.
         
         Args:
             items: (any,)
@@ -134,26 +232,31 @@ class Linque(object):
         return Linque(result, self._evaluate)
     
     
-    def contains(self, item):
+    def contains(self, value, key=None):
         """
-        Determines whether current sequence contains specified item by using
-        default comparer. This call partially evaluates current sequence.
+        Determines whether current sequence contains specified item or value by using
+        default comparer or specified item's key.
         
         Args:
-            item: any
-                Item to check.
+            value: any
+                Item or value to check.
+            
+            key: callable or None
+                Item's key selector.
         
         Returns:
             bool
         """
         
-        return item in self
+        if key is not None:
+            return value in (key(d) for d in self)
+        
+        return value in self
     
     
     def count(self, condition=None):
         """
         Returns number of items in current sequence satisfying given condition.
-        This call fully evaluates current sequence.
         
         Args:
             condition: callable or None
@@ -166,46 +269,32 @@ class Linque(object):
         return iters.count(self, condition)
     
     
-    def distinct(self):
+    def distinct(self, key=None):
         """
         Produces new sequence by selecting distinct items from current sequence
-        using default comparer. First occurrence of each item is used. This call
-        does not evaluate current sequence.
-        
-        Returns:
-            Linque
-        """
-        
-        result = iters.distinct(self)
-        
-        return Linque(result, self._evaluate)
-    
-    
-    def distinct_by(self, key):
-        """
-        Produces new sequence by selecting distinct items from current sequence
-        using specified item's key. First occurrence of each item is used. This
-        call does not evaluate current sequence.
+        using default comparer or specified item's key. First occurrence of each
+        item is used.
         
         Args:
-            key: callable
+            key: callable or None
                 Item's key selector.
         
         Returns:
             Linque
         """
         
-        result = iters.distinct_by(self, key)
+        result = iters.distinct(self, key)
         
         return Linque(result, self._evaluate)
     
     
     def each(self, action):
         """
-        Applies specified function to every item in current sequence. Note that
-        this does not replace the items, so it works only with objects. Any
-        return value of given function is ignored. This call fully evaluates
-        current sequence.
+        Applies specified function to every item in current sequence. Any
+        return value of given function is ignored. Since this call fully
+        evaluates current sequence, depending on the source, the items may no
+        longer be available. This may or may not be desired behavior. Consider
+        calling '.evaluate()' before calling this method.
         
         Args:
             action: callable
@@ -215,8 +304,6 @@ class Linque(object):
             Linque
         """
         
-        self.evaluate()
-        
         for item in self:
             action(item)
         
@@ -226,7 +313,7 @@ class Linque(object):
     def enumerate(self):
         """
         Produces new sequence by enumerating items of current sequence into
-        (index, item) pairs. This call does not evaluate current sequence.
+        (index, item) pairs.
         
         Returns:
             Linque
@@ -240,80 +327,44 @@ class Linque(object):
     def evaluate(self):
         """
         Evaluates all the iterators in current sequence and stores items as
-        internal list. This method is essential if current Linque instance
-        should be reused.
+        internal list.
         
         Returns:
             Linque
         """
         
-        if not isinstance(self._items, (list, tuple)):
-            self._items = list(self)
+        if not isinstance(self._source, (list, tuple)):
+            self._source = list(self)
         
         return self
     
     
-    def exclude(self, items):
+    def exclude(self, items, key=None):
         """
         Produces new sequence by excluding specified items from current sequence
-        using default comparer. This call does not evaluate current sequence.
-        
-        Args:
-            items: (any,)
-                Items to exclude.
-        
-        Returns:
-            Linque
-        """
-        
-        result = iters.exclude(self, items)
-        
-        return Linque(result, self._evaluate)
-    
-    
-    def exclude_by(self, items, key):
-        """
-        Produces new sequence by excluding specified items from current sequence
-        using selected item's key. This call does not evaluate current sequence.
+        using default comparer or specified item's key.
         
         Args:
             items: (any,)
                 Items to exclude.
             
-            key: callable
+            key: callable or None
                 Item's key selector.
         
         Returns:
             Linque
         """
         
-        result = iters.exclude_by(self, items, key)
+        result = iters.exclude(self, items, key)
         
         return Linque(result, self._evaluate)
     
     
-    def first(self, condition=None):
+    def first(self, condition=None, default=iters.UNDEFINED):
         """
         Returns the first item in current sequence that satisfies specified
-        condition or raises error if no item found. This call partially
-        evaluates current sequence.
-        
-        Args:
-            condition: callable or None
-                Condition to test.
-        
-        Returns:
-            any
-        """
-        
-        return iters.first(self, condition)
-    
-    
-    def first_or_default(self, condition=None, default=None):
-        """
-        Returns the first item in current sequence that satisfies specified
-        condition or specified default value if no item found. This call
-        partially evaluates current sequence.
+        condition or raises error if no item found and no default value is
+        provided.
         
         Args:
             condition: callable or None
@@ -326,13 +377,13 @@ class Linque(object):
             any
         """
         
-        return iters.first_or_default(self, condition, default)
+        return iters.first(self, condition, default)
     
     
     def flatten(self, selector=None):
         """
         Produces new sequence by selecting and flattening items data using
-        specified selector. This call does not evaluate current sequence.
+        specified selector.
         
         Args:
             selector: callable
@@ -350,105 +401,55 @@ class Linque(object):
         return Linque(result, self._evaluate)
     
     
-    def group(self):
+    def group(self, key=None):
         """
         Produces new sequence by grouping items of current sequence according to
-        default comparer and creates result values as (key, group) pairs.
-        This call fully evaluates current sequence.
-        
-        Returns:
-            Linque
-        """
-        
-        groups = iters.group(self)
-        result = [(k, Linque(g, self._evaluate)) for k, g in groups]
-        
-        return Linque(result, self._evaluate)
-    
-    
-    def group_by(self, key):
-        """
-        Produces new sequence by grouping items of current sequence according to
-        specified key selector and creates result values as (key, group) pairs.
-        This call fully evaluates current sequence.
+        default comparer or specified item's key and creates result values as
+        (key, group) pairs.
         
         Args:
-            key: callable
+            key: callable or None
                 Item's key selector.
         
         Returns:
             Linque
         """
         
-        groups = iters.group_by(self, key)
-        result = [(k, Linque(g, self._evaluate)) for k, g in groups]
+        def source():
+            for item in iters.group(self, key):
+                yield item
+        
+        result = ((k, Linque(g, self._evaluate)) for k, g in source())
         
         return Linque(result, self._evaluate)
     
     
-    def intersect(self, items):
+    def intersect(self, items, key=None):
         """
         Produces new sequence of shared unique items from current sequence and
-        given items by using default comparer. This call does not evaluate
-        current sequence.
-        
-        Args:
-            items: (any,)
-                Items to intersect.
-        
-        Returns:
-            Linque
-        """
-        
-        result = iters.intersect(self, items)
-        
-        return Linque(result, self._evaluate)
-    
-    
-    def intersect_by(self, items, key):
-        """
-        Produces new sequence of shared unique items from current sequence and
-        given items by using selected item's key. This call does not evaluate
-        current sequence.
+        given items by using default comparer or specified item's key.
         
         Args:
             items: (any,)
                 Items to intersect.
             
-            key: callable
+            key: callable or None
                 Item's key selector.
         
         Returns:
             Linque
         """
         
-        result = iters.intersect_by(self, items, key)
+        result = iters.intersect(self, items, key)
         
         return Linque(result, self._evaluate)
     
     
-    def last(self, condition=None):
+    def last(self, condition=None, default=iters.UNDEFINED):
         """
         Returns the last item in current sequence that satisfies specified
-        condition or raises error if no item found. This call fully evaluates
-        current sequence.
-        
-        Args:
-            condition: callable or None
-                Condition to test.
-        
-        Returns:
-            any
-        """
-        
-        return iters.last(self, condition)
-    
-    
-    def last_or_default(self, condition=None, default=None):
-        """
-        Returns the last item in current sequence that satisfies specified
-        condition or specified default value if no item found. This call
-        fully evaluates current sequence.
+        condition or raises error if no item found and no default value is
+        provided.
         
         Args:
             condition: callable or None
@@ -461,13 +462,13 @@ class Linque(object):
             any
         """
         
-        return iters.last_or_default(self, condition, default)
+        return iters.last(self, condition, default)
     
     
-    def max(self, selector=None):
+    def maximum(self, selector=None):
         """
         Returns maximum value in current sequence by specified items data
-        selector. This call fully evaluates current sequence.
+        selector.
         
         Args:
             selector: callable
@@ -483,27 +484,26 @@ class Linque(object):
         return max(selector(d) for d in self)
     
     
-    def max_by(self, key):
+    def max(self, key=None):
         """
         Returns item having maximum value in current sequence by using default
-        comparer or specified item's key. This call fully evaluates current
-        sequence.
+        comparer or specified item's key.
         
         Args:
-            key: callable
+            key: callable or None
                 Item's key selector.
         
         Returns:
             any
         """
         
-        return max(self, key=key)
+        return max(self, key=key) if key is not None else max(self)
     
     
     def mean(self, selector=None):
         """
         Returns average value of current sequence by specified items data
-        selector. This call fully evaluates current sequence.
+        selector.
         
         Args:
             selector: callable
@@ -522,7 +522,7 @@ class Linque(object):
     def median(self, selector=None):
         """
         Returns median value of current sequence by specified items data
-        selector. This call fully evaluates current sequence.
+        selector.
         
         Args:
             selector: callable
@@ -538,10 +538,10 @@ class Linque(object):
         return statistics.median(selector(d) for d in self)
     
     
-    def min(self, selector=None):
+    def minimum(self, selector=None):
         """
         Returns minimum value in current sequence by specified items data
-        selector. This call fully evaluates current sequence.
+        selector.
         
         Args:
             selector: callable
@@ -557,33 +557,91 @@ class Linque(object):
         return min(selector(d) for d in self)
     
     
-    def min_by(self, key):
+    def min(self, key=None):
         """
         Returns item having minimum value in current sequence by using default
-        comparer or specified item's key. This call fully evaluates current
-        sequence.
+        comparer or specified item's key.
         
         Args:
-            key: callable
+            key: callable or None
                 Item's key selector.
         
         Returns:
             any
         """
         
-        return min(self, key=key)
+        return min(self, key=key) if key is not None else min(self)
     
     
-    def reverse(self):
+    def rank(self, key=None, method='average', reverse=False):
         """
-        Produces new sequence by inverting order of items in current sequence.
-        This call fully evaluates current sequence.
+        Provides 1-based rank for each item of current sequence by using default
+        comparer or specified item's key. The ties are resolved according to
+        selected method.
+        
+        Args:
+            key: callable or None
+                Item's key selector.
+            
+            method: str
+                Method used to assign ranks to tied items.
+                    'average' - tied values are assigned by their average rank
+                    'min' - tied values are assigned by their minimum rank
+                    'max' - tied values are assigned by their maximum rank
+                    'dense' - like 'min' but without rank gaps
+                    'ordinal' - distinct rank for all values
+            
+            reverse: bool
+                If set to True, sorting is reversed.
         
         Returns:
             Linque
         """
         
-        result = reversed([d for d in self])
+        def source():
+            for item in iters.rank(self, key, method=method, reverse=reverse):
+                yield item
+        
+        result = (d for d in source())
+        
+        return Linque(result, self._evaluate)
+    
+    
+    def reverse(self):
+        """
+        Produces new sequence by inverting order of items in current sequence.
+        
+        Returns:
+            Linque
+        """
+        
+        def source():
+            for item in reversed(list(self)):
+                yield item
+        
+        result = (d for d in source())
+        
+        return Linque(result, self._evaluate)
+    
+    
+    def sample(self, count):
+        """
+        Produces new sequence by randomly sampling number of items from current
+        sequence. Each item can be selected only once.
+        
+        Args:
+            count: int
+                Number of items to sample.
+        
+        Returns:
+            Linque
+        """
+        
+        def source():
+            for item in random.sample(list(self), count):
+                yield item
+        
+        result = (d for d in source())
         
         return Linque(result, self._evaluate)
     
@@ -591,7 +649,6 @@ class Linque(object):
     def select(self, selector):
         """
         Produces new sequence by selecting items data by specified selector.
-        This call does not evaluate current sequence.
         
         Args:
             selector: callable
@@ -609,7 +666,7 @@ class Linque(object):
     def select_many(self, selector=None):
         """
         Produces new sequence by selecting and flattening items data using
-        specified selector. This call does not evaluate current sequence.
+        specified selector.
         
         Args:
             selector: callable
@@ -627,28 +684,49 @@ class Linque(object):
         return Linque(result, self._evaluate)
     
     
-    def single(self, condition=None):
+    def shuffle(self):
+        """
+        Produces new sequence by randomly shuffling items from current sequence.
+        
+        Returns:
+            Linque
+        """
+        
+        def source():
+            items = list(self)
+            random.shuffle(items)
+            for item in items:
+                yield item
+        
+        result = (d for d in source())
+        
+        return Linque(result, self._evaluate)
+    
+    
+    def single(self, condition=None, default=iters.UNDEFINED):
         """
         Returns the single item in current sequence that satisfies specified
-        condition or raises error if none or more items found. This call fully
-        evaluates current sequence.
+        condition or raises error if more items found. If no item is found,
+        returns default value if provided or raises error.
         
         Args:
             condition: callable
                 Condition to test.
+            
+            default: any
+                Default value.
         
         Returns:
             any
         """
         
-        return iters.single(self, condition)
+        return iters.single(self, condition, default)
     
     
     def skip(self, count):
         """
         Produces new sequence by bypassing specified number of items in current
-        sequence and returns the remaining items. This call partially evaluates
-        current sequence.
+        sequence and returns the remaining items.
         
         Args:
             count: int
@@ -681,68 +759,30 @@ class Linque(object):
         return Linque(result, self._evaluate)
     
     
-    def sort(self):
+    def sort(self, key=None, reverse=False):
         """
-        Produces new sequence by sorting elements of current sequence in
-        ascending order by using default comparer. This call fully evaluates
-        current sequence.
-        
-        Returns:
-            Linque
-        """
-        
-        result = sorted(self)
-        
-        return Linque(result, self._evaluate)
-    
-    
-    def sort_desc(self):
-        """
-        Sorts elements of current sequence in descending order by using default
-        comparer. This call fully evaluates current sequence.
-        
-        Returns:
-            Linque
-        """
-        
-        result = sorted(self, reverse=True)
-        
-        return Linque(result, self._evaluate)
-    
-    
-    def sort_by(self, key):
-        """
-        Produces new sequence by sorting elements of current sequence in
-        ascending order by using selected item's key. This call fully evaluates
-        current sequence.
+        Produces new sequence by sorting elements of current sequence by using
+        default comparer or specified item's key. If the key provides multiple
+        columns, the sorting direction can be specified for each individual
+        column.
         
         Args:
-            key: callable
+            key: callable or None
                 Item's key selector.
+            
+            reverse: bool
+                If set to True, sorting is reversed. This flag can be specified
+                independently foreach key column.
         
         Returns:
             Linque
         """
         
-        result = sorted(self, key=key)
+        def source():
+            for item in iters.multisort(self, key=key, reverse=reverse):
+                yield item
         
-        return Linque(result, self._evaluate)
-    
-    
-    def sort_by_desc(self, key):
-        """
-        Sorts elements of current sequence in descending order by using selected
-        item's key. This call fully evaluates current sequence.
-        
-        Args:
-            key: callable
-                Item's key selector.
-        
-        Returns:
-            Linque
-        """
-        
-        result = sorted(self, key=key, reverse=True)
+        result = (d for d in source())
         
         return Linque(result, self._evaluate)
     
@@ -750,7 +790,7 @@ class Linque(object):
     def sum(self, selector=None):
         """
         Returns summed value in current sequence by specified items data
-        selector. This call fully evaluates current sequence.
+        selector.
         
         Args:
             selector: callable
@@ -769,8 +809,7 @@ class Linque(object):
     def take(self, count):
         """
         Produces new sequence by selecting specified number of contiguous items
-        from the start of current sequence. This call partially evaluates
-        current sequence.
+        from the start of current sequence.
         
         Args:
             count: int
@@ -788,8 +827,7 @@ class Linque(object):
     def take_while(self, condition):
         """
         Produces new sequence by selecting items from current sequence as long
-        as specified condition is true. This call partially evaluates current
-        sequence.
+        as specified condition is true.
         
         Args:
             condition: callable
@@ -806,8 +844,7 @@ class Linque(object):
     
     def to_dict(self, key, value=lambda d: d):
         """
-        Evaluates items into dictionary. This call fully evaluates current
-        sequence.
+        Evaluates items into dictionary.
         
         Args:
             key: callable
@@ -833,7 +870,7 @@ class Linque(object):
     
     def to_list(self):
         """
-        Evaluate items into list. This call fully evaluates current sequence.
+        Evaluate items into list.
         
         Returns:
             list
@@ -844,7 +881,7 @@ class Linque(object):
     
     def to_set(self):
         """
-        Evaluate items into set. This call fully evaluates current sequence.
+        Evaluate items into set.
         
         Returns:
             set
@@ -855,7 +892,7 @@ class Linque(object):
     
     def to_tuple(self):
         """
-        Evaluate items into tuple. This call fully evaluates current sequence.
+        Evaluate items into tuple.
         
         Returns:
             tuple
@@ -864,51 +901,30 @@ class Linque(object):
         return tuple(self)
     
     
-    def union(self, items):
+    def union(self, items, key=None):
         """
         Produces new sequence of unique items from current sequence and given
-        items by using default comparer. This call does not evaluate current
-        sequence.
-        
-        Args:
-            items: (any,)
-                Items to union.
-        
-        Returns:
-            Linque
-        """
-        
-        result = iters.union(self, items)
-        
-        return Linque(result, self._evaluate)
-    
-    
-    def union_by(self, items, key):
-        """
-        Produces new sequence of unique items from current sequence and given
-        items by using selected item's key. This call does not evaluate current
-        sequence.
+        items by using default comparer or specified item's key.
         
         Args:
             items: (any,)
                 Items to union.
             
-            key: callable
+            key: callable or None
                 Item's key selector.
         
         Returns:
             Linque
         """
         
-        result = iters.union_by(self, items, key)
+        result = iters.union(self, items, key)
         
         return Linque(result, self._evaluate)
     
     
     def where(self, condition):
         """
-        Produces new sequence by selecting items by specified predicate. This
-        call does not evaluate current sequence.
+        Produces new sequence by selecting items by specified predicate.
         
         Args:
             condition: callable
@@ -926,8 +942,7 @@ class Linque(object):
     def zip(self, *sequences):
         """
         Produces new sequence by merging given sequences with current sequence
-        as long as there are items available in all sequences. This call does
-        not evaluate current sequence.
+        as long as there are items available in all sequences.
         
         Args:
             sequences: ((any,),)
