@@ -53,6 +53,107 @@ class TestCase(unittest.TestCase):
         self.assertTrue(linq.any(lambda d: d > 5))
     
     
+    def test_argmax(self):
+        """Tests whether argmax works correctly."""
+        
+        data = (3, 1, 2, 0, 9, 7, 8)
+        
+        linq = linque.Linque(data)
+        self.assertEqual(linq.argmax(), 4)
+        
+        linq = linque.Linque(d for d in data)
+        self.assertEqual(linq.argmax(), 4)
+        
+        data = ((0, 3), (1, 1), (2, 2), (3, 0), (4, 9), (5, 7), (6, 8))
+        
+        linq = linque.Linque(data)
+        self.assertEqual(linq.argmax(lambda d: d[1]), 4)
+        
+        linq = linque.Linque(d for d in data)
+        self.assertEqual(linq.argmax(lambda d: d[1]), 4)
+    
+    
+    def test_argmin(self):
+        """Tests whether argmin works correctly."""
+        
+        data = (3, 1, 2, 0, 9, 7, 8)
+        
+        linq = linque.Linque(data)
+        self.assertEqual(linq.argmin(), 3)
+        
+        linq = linque.Linque(d for d in data)
+        self.assertEqual(linq.argmin(), 3)
+        
+        data = ((0, 3), (1, 1), (2, 2), (3, 0), (4, 9), (5, 7), (6, 8))
+        
+        linq = linque.Linque(data)
+        self.assertEqual(linq.argmin(lambda d: d[1]), 3)
+        
+        linq = linque.Linque(d for d in data)
+        self.assertEqual(linq.argmin(lambda d: d[1]), 3)
+    
+    
+    def test_argsort(self):
+        """Tests whether argsort works correctly."""
+        
+        data = (3, 1, 2)
+        
+        linq = linque.Linque(data)
+        self.assertEqual(linq.argsort().to_tuple(), (1, 2, 0))
+        self.assertEqual(linq.argsort(reverse=True).to_tuple(), (0, 2, 1))
+        
+        linq = linque.Linque(d for d in data)
+        self.assertEqual(linq.argsort().to_tuple(), (1, 2, 0))
+        
+        linq = linque.Linque(d for d in data)
+        self.assertEqual(linq.argsort(reverse=True).to_tuple(), (0, 2, 1))
+        
+        data = ((2, 3), (1, 1), (3, 2))
+        
+        linq = linque.Linque(data)
+        self.assertEqual(linq.argsort(lambda d: d[1]).to_tuple(), (1, 2, 0))
+        self.assertEqual(linq.argsort(lambda d: d[1], reverse=True).to_tuple(), (0, 2, 1))
+        
+        linq = linque.Linque(d for d in data)
+        self.assertEqual(linq.argsort(lambda d: d[1]).to_tuple(), (1, 2, 0))
+        
+        linq = linque.Linque(d for d in data)
+        self.assertEqual(linq.argsort(lambda d: d[1], reverse=True).to_tuple(), (0, 2, 1))
+    
+    
+    def test_choice(self):
+        """Tests whether choice works correctly."""
+        
+        data = (0, 1, 2, 3, 4)
+        
+        linq = linque.Linque(data)
+        self.assertTrue(type(linq.choice()) == int)
+        self.assertTrue(type(linq.choice(weights=[5, 5, 10, 5, 5])) == int)
+        
+        linq = linque.Linque(d for d in data)
+        self.assertTrue(type(linq.choice()) == int)
+        
+        linq = linque.Linque(d for d in data)
+        self.assertTrue(type(linq.choice(weights=[5, 5, 10, 5, 5])) == int)
+    
+    
+    def test_choices(self):
+        """Tests whether choices works correctly."""
+        
+        data = (0, 1, 2, 3, 4)
+        count = 10
+        
+        linq = linque.Linque(data)
+        self.assertTrue(len(linq.choices(count).to_list()) == count)
+        self.assertTrue(len(linq.choices(count, weights=[5, 5, 10, 5, 5]).to_list()) == count)
+        
+        linq = linque.Linque(d for d in data)
+        self.assertTrue(len(linq.choices(count, ).to_list()) == count)
+        
+        linq = linque.Linque(d for d in data)
+        self.assertTrue(len(linq.choices(count, weights=[5, 5, 10, 5, 5]).to_list()) == count)
+    
+    
     def test_concat(self):
         """Tests whether concat works correctly."""
         
@@ -82,6 +183,16 @@ class TestCase(unittest.TestCase):
         
         linq = linque.Linque(d for d in data)
         self.assertFalse(linq.contains((1, -10)))
+        
+        linq = linque.Linque(data)
+        self.assertTrue(linq.contains(10, lambda d: d[1]))
+        self.assertFalse(linq.contains(-10, lambda d: d[1]))
+        
+        linq = linque.Linque(d for d in data)
+        self.assertTrue(linq.contains(10, lambda d: d[1]))
+        
+        linq = linque.Linque(d for d in data)
+        self.assertFalse(linq.contains(-10, lambda d: d[1]))
     
     
     def test_chunk(self):
@@ -124,18 +235,12 @@ class TestCase(unittest.TestCase):
         
         linq = linque.Linque(d for d in data)
         self.assertEqual(linq.distinct().to_tuple(), ((0, 1), (0, 2), (1, 1), (1, 2)))
-    
-    
-    def test_distinct_by(self):
-        """Tests whether distinct_by works correctly."""
-        
-        data = ((0, 1), (0, 1), (0, 2), (1, 1), (1, 2))
         
         linq = linque.Linque(data)
-        self.assertEqual(linq.distinct_by(lambda d: d[1]).to_tuple(), ((0, 1), (0, 2)))
+        self.assertEqual(linq.distinct(lambda d: d[1]).to_tuple(), ((0, 1), (0, 2)))
         
         linq = linque.Linque(d for d in data)
-        self.assertEqual(linq.distinct_by(lambda d: d[1]).to_tuple(), ((0, 1), (0, 2)))
+        self.assertEqual(linq.distinct(lambda d: d[1]).to_tuple(), ((0, 1), (0, 2)))
     
     
     def test_each(self):
@@ -148,9 +253,10 @@ class TestCase(unittest.TestCase):
         
         linq = linque.Linque(data)
         self.assertEqual(linq.each(action).to_tuple(), ([0, '0'], [1, '1'], [2, '2'], [3, '3'], [4, '4']))
+        self.assertEqual(data, ([0, '0'], [1, '1'], [2, '2'], [3, '3'], [4, '4']))
         
         linq = linque.Linque(d for d in data)
-        self.assertEqual(linq.each(action).to_tuple(), ([0, '0'], [1, '1'], [2, '2'], [3, '3'], [4, '4']))
+        self.assertEqual(linq.each(action).to_tuple(), ())
     
     
     def test_enumerate(self):
@@ -192,21 +298,14 @@ class TestCase(unittest.TestCase):
         linq = linque.Linque(d for d in data1)
         items2 = (d for d in data2)
         self.assertEqual(linq.exclude(items2).to_tuple(), ((0, 2), (0, 3), (0, 4)))
-    
-    
-    def test_exclude_by(self):
-        """Tests whether exclude_by works correctly."""
-        
-        data1 = ((0, 1), (0, 1), (0, 2), (1, 2), (0, 3), (0, 4))
-        data2 = ((0, 1), (1, 2), (1, 2), (1, 3))
         
         linq = linque.Linque(data1)
         items2 = data2
-        self.assertEqual(linq.exclude_by(items2, lambda d: d[1]).to_tuple(), ((0, 4),))
+        self.assertEqual(linq.exclude(items2, lambda d: d[1]).to_tuple(), ((0, 4),))
         
         linq = linque.Linque(d for d in data1)
         items2 = (d for d in data2)
-        self.assertEqual(linq.exclude_by(items2, lambda d: d[1]).to_tuple(), ((0, 4),))
+        self.assertEqual(linq.exclude(items2, lambda d: d[1]).to_tuple(), ((0, 4),))
     
     
     def test_first(self):
@@ -226,6 +325,12 @@ class TestCase(unittest.TestCase):
         linq = linque.Linque(d for d in data)
         self.assertEqual(linq.first(lambda d: d > 4), 5)
         
+        linq = linque.Linque(data)
+        self.assertEqual(linq.first(lambda d: d > 10, -1), -1)
+        
+        linq = linque.Linque(d for d in data)
+        self.assertEqual(linq.first(lambda d: d > 10, -1), -1)
+        
         linq = linque.Linque(data)
         with self.assertRaises(StopIteration):
             linq.first(lambda d: d > 10)
@@ -235,30 +340,6 @@ class TestCase(unittest.TestCase):
             linq.first(lambda d: d > 10)
     
     
-    def test_first_or_default(self):
-        """Tests whether first_or_default works correctly."""
-        
-        data = (0, 1, 2, 3, 4, 5, 6, 7, 8, 9)
-        
-        linq = linque.Linque(data)
-        self.assertEqual(linq.first_or_default(), 0)
-        
-        linq = linque.Linque(d for d in data)
-        self.assertEqual(linq.first_or_default(), 0)
-        
-        linq = linque.Linque(data)
-        self.assertEqual(linq.first_or_default(lambda d: d > 4), 5)
-        
-        linq = linque.Linque(d for d in data)
-        self.assertEqual(linq.first_or_default(lambda d: d > 4), 5)
-        
-        linq = linque.Linque(data)
-        self.assertEqual(linq.first_or_default(lambda d: d > 10, -1), -1)
-        
-        linq = linque.Linque(d for d in data)
-        self.assertEqual(linq.first_or_default(lambda d: d > 10, -1), -1)
-    
-    
     def test_flatten(self):
         """Tests whether flatten works correctly."""
         
@@ -287,20 +368,14 @@ class TestCase(unittest.TestCase):
             (0, 1): ((0, 1), (0, 1)),
             (0, 2): ((0, 2),),
             (1, 1): ((1, 1),)})
-    
-    
-    def test_group_by(self):
-        """Tests whether group_by works correctly."""
-        
-        data = ((0, 1), (0, 1), (0, 2), (1, 1))
         
         linq = linque.Linque(data)
-        self.assertEqual(linq.group_by(lambda d: d[1]).to_dict(lambda d: d[0], lambda d: d[1].to_tuple()), {
+        self.assertEqual(linq.group(lambda d: d[1]).to_dict(lambda d: d[0], lambda d: d[1].to_tuple()), {
             1: ((0, 1), (0, 1), (1, 1)),
             2: ((0, 2),)})
         
         linq = linque.Linque(d for d in data)
-        self.assertEqual(linq.group_by(lambda d: d[1]).to_dict(lambda d: d[0], lambda d: d[1].to_tuple()), {
+        self.assertEqual(linq.group(lambda d: d[1]).to_dict(lambda d: d[0], lambda d: d[1].to_tuple()), {
             1: ((0, 1), (0, 1), (1, 1)),
             2: ((0, 2),)})
     
@@ -318,21 +393,14 @@ class TestCase(unittest.TestCase):
         linq = linque.Linque(d for d in data1)
         items2 = (d for d in data2)
         self.assertEqual(linq.intersect(items2).to_tuple(), ((0, 1), (1, 2)))
-    
-    
-    def test_intersect_by(self):
-        """Tests whether intersect_by works correctly."""
-        
-        data1 = ((0, 1), (0, 1), (0, 2), (1, 2))
-        data2 = ((0, 1), (1, 2), (1, 2), (0, 3))
         
         linq = linque.Linque(data1)
         items2 = data2
-        self.assertEqual(linq.intersect_by(items2, lambda d: d[1]).to_tuple(), ((0, 1), (0, 2)))
+        self.assertEqual(linq.intersect(items2, lambda d: d[1]).to_tuple(), ((0, 1), (0, 2)))
         
         linq = linque.Linque(d for d in data1)
         items2 = (d for d in data2)
-        self.assertEqual(linq.intersect_by(items2, lambda d: d[1]).to_tuple(), ((0, 1), (0, 2)))
+        self.assertEqual(linq.intersect(items2, lambda d: d[1]).to_tuple(), ((0, 1), (0, 2)))
     
     
     def test_last(self):
@@ -352,6 +420,12 @@ class TestCase(unittest.TestCase):
         linq = linque.Linque(d for d in data)
         self.assertEqual(linq.last(lambda d: d > 4), 6)
         
+        linq = linque.Linque(data)
+        self.assertEqual(linq.last(lambda d: d > 10, -1), -1)
+        
+        linq = linque.Linque(d for d in data)
+        self.assertEqual(linq.last(lambda d: d > 10, -1), -1)
+        
         linq = linque.Linque(data)
         with self.assertRaises(StopIteration):
             linq.last(lambda d: d > 10)
@@ -361,60 +435,42 @@ class TestCase(unittest.TestCase):
             linq.last(lambda d: d > 10)
     
     
-    def test_last_or_default(self):
-        """Tests whether last_or_default works correctly."""
-        
-        data = (0, 1, 2, 3, 4, 5, 4, 5, 6, 1)
-        
-        linq = linque.Linque(data)
-        self.assertEqual(linq.last_or_default(), 1)
-        
-        linq = linque.Linque(d for d in data)
-        self.assertEqual(linq.last_or_default(), 1)
-        
-        linq = linque.Linque(data)
-        self.assertEqual(linq.last_or_default(lambda d: d > 4), 6)
-        
-        linq = linque.Linque(d for d in data)
-        self.assertEqual(linq.last_or_default(lambda d: d > 4), 6)
-        
-        linq = linque.Linque(data)
-        self.assertEqual(linq.last_or_default(lambda d: d > 10, -1), -1)
-        
-        linq = linque.Linque(d for d in data)
-        self.assertEqual(linq.last_or_default(lambda d: d > 10, -1), -1)
-    
-    
     def test_max(self):
         """Tests whether max works correctly."""
         
+        data = ((0, 0), (1, 100), (20, 20), (3, 30), (4, 40))
+        
+        linq = linque.Linque(data)
+        self.assertEqual(linq.max(), (20, 20))
+        
+        linq = linque.Linque(d for d in data)
+        self.assertEqual(linq.max(), (20, 20))
+        
+        linq = linque.Linque(data)
+        self.assertEqual(linq.max(lambda d: d[1]), (1, 100))
+        
+        linq = linque.Linque(d for d in data)
+        self.assertEqual(linq.max(lambda d: d[1]), (1, 100))
+    
+    
+    def test_maximum(self):
+        """Tests whether max works correctly."""
+        
         data = (0, 1, 2, 3, 4)
         
         linq = linque.Linque(data)
-        self.assertEqual(linq.max(), 4)
+        self.assertEqual(linq.maximum(), 4)
         
         linq = linque.Linque(d for d in data)
-        self.assertEqual(linq.max(), 4)
+        self.assertEqual(linq.maximum(), 4)
         
         data = ((0, 0), (1, 10), (2, 20), (3, 30), (4, 40))
         
         linq = linque.Linque(data)
-        self.assertEqual(linq.max(lambda d: d[1]), 40)
+        self.assertEqual(linq.maximum(lambda d: d[1]), 40)
         
         linq = linque.Linque(d for d in data)
-        self.assertEqual(linq.max(lambda d: d[1]), 40)
-    
-    
-    def test_max_by(self):
-        """Tests whether max_by works correctly."""
-        
-        data = ((0, 0), (1, 10), (2, 20), (3, 30), (4, 40))
-        
-        linq = linque.Linque(data)
-        self.assertEqual(linq.max_by(lambda d: d[1]), (4, 40))
-        
-        linq = linque.Linque(d for d in data)
-        self.assertEqual(linq.max_by(lambda d: d[1]), (4, 40))
+        self.assertEqual(linq.maximum(lambda d: d[1]), 40)
     
     
     def test_mean(self):
@@ -460,33 +516,91 @@ class TestCase(unittest.TestCase):
     def test_min(self):
         """Tests whether min works correctly."""
         
+        data = ((0, 0), (1, -100), (-2, -20), (3, -30), (4, -40))
+        
+        linq = linque.Linque(data)
+        self.assertEqual(linq.min(), (-2, -20))
+        
+        linq = linque.Linque(d for d in data)
+        self.assertEqual(linq.min(), (-2, -20))
+        
+        linq = linque.Linque(data)
+        self.assertEqual(linq.min(lambda d: d[1]), (1, -100))
+        
+        linq = linque.Linque(d for d in data)
+        self.assertEqual(linq.min(lambda d: d[1]), (1, -100))
+    
+    
+    def test_minimum(self):
+        """Tests whether min works correctly."""
+        
         data = (0, -1, -2, -3, -4)
         
         linq = linque.Linque(data)
-        self.assertEqual(linq.min(), -4)
+        self.assertEqual(linq.minimum(), -4)
         
         linq = linque.Linque(d for d in data)
-        self.assertEqual(linq.min(), -4)
+        self.assertEqual(linq.minimum(), -4)
         
         data = ((0, 0), (1, -10), (2, -20), (3, -30), (4, -40))
         
         linq = linque.Linque(data)
-        self.assertEqual(linq.min(lambda d: d[1]), -40)
+        self.assertEqual(linq.minimum(lambda d: d[1]), -40)
         
         linq = linque.Linque(d for d in data)
-        self.assertEqual(linq.min(lambda d: d[1]), -40)
+        self.assertEqual(linq.minimum(lambda d: d[1]), -40)
     
     
-    def test_min_by(self):
-        """Tests whether min_by works correctly."""
+    def test_rank(self):
+        """Tests whether rank works correctly."""
         
-        data = ((0, 0), (1, -10), (2, -20), (3, -30), (4, -40))
+        data = (0, 2, 3, 2)
         
         linq = linque.Linque(data)
-        self.assertEqual(linq.min_by(lambda d: d[1]), (4, -40))
+        self.assertEqual(linq.rank(method='average').to_tuple(), (1, 2.5, 4, 2.5))
+        self.assertEqual(linq.rank(method='min').to_tuple(), (1, 2, 4, 2))
+        self.assertEqual(linq.rank(method='max').to_tuple(), (1, 3, 4, 3))
+        self.assertEqual(linq.rank(method='dense').to_tuple(), (1, 2, 3, 2))
+        self.assertEqual(linq.rank(method='ordinal').to_tuple(), (1, 2, 4, 3))
         
         linq = linque.Linque(d for d in data)
-        self.assertEqual(linq.min_by(lambda d: d[1]), (4, -40))
+        self.assertEqual(linq.rank(method='average').to_tuple(), (1, 2.5, 4, 2.5))
+        
+        linq = linque.Linque(d for d in data)
+        self.assertEqual(linq.rank(method='min').to_tuple(), (1, 2, 4, 2))
+        
+        linq = linque.Linque(d for d in data)
+        self.assertEqual(linq.rank(method='max').to_tuple(), (1, 3, 4, 3))
+        
+        linq = linque.Linque(d for d in data)
+        self.assertEqual(linq.rank(method='dense').to_tuple(), (1, 2, 3, 2))
+        
+        linq = linque.Linque(d for d in data)
+        self.assertEqual(linq.rank(method='ordinal').to_tuple(), (1, 2, 4, 3))
+        
+        data = ((2, 0), (3, 2), (2, 3), (0, 2))
+        
+        linq = linque.Linque(data)
+        self.assertEqual(linq.rank(lambda d: d[1], method='average').to_tuple(), (1, 2.5, 4, 2.5))
+        self.assertEqual(linq.rank(lambda d: d[1], method='min').to_tuple(), (1, 2, 4, 2))
+        self.assertEqual(linq.rank(lambda d: d[1], method='max').to_tuple(), (1, 3, 4, 3))
+        self.assertEqual(linq.rank(lambda d: d[1], method='dense').to_tuple(), (1, 2, 3, 2))
+        self.assertEqual(linq.rank(lambda d: d[1], method='ordinal').to_tuple(), (1, 2, 4, 3))
+        
+        linq = linque.Linque(d for d in data)
+        self.assertEqual(linq.rank(lambda d: d[1], method='average').to_tuple(), (1, 2.5, 4, 2.5))
+        
+        linq = linque.Linque(d for d in data)
+        self.assertEqual(linq.rank(lambda d: d[1], method='min').to_tuple(), (1, 2, 4, 2))
+        
+        linq = linque.Linque(d for d in data)
+        self.assertEqual(linq.rank(lambda d: d[1], method='max').to_tuple(), (1, 3, 4, 3))
+        
+        linq = linque.Linque(d for d in data)
+        self.assertEqual(linq.rank(lambda d: d[1], method='dense').to_tuple(), (1, 2, 3, 2))
+        
+        linq = linque.Linque(d for d in data)
+        self.assertEqual(linq.rank(lambda d: d[1], method='ordinal').to_tuple(), (1, 2, 4, 3))
     
     
     def test_reverse(self):
@@ -501,6 +615,19 @@ class TestCase(unittest.TestCase):
         self.assertEqual(linq.reverse().to_tuple(), (9, 8, 7, 6, 5, 4, 3, 2, 1, 0))
     
     
+    def test_sample(self):
+        """Tests whether sample works correctly."""
+        
+        data = (0, 1, 2, 3, 4)
+        count = 3
+        
+        linq = linque.Linque(data)
+        self.assertTrue(len(linq.sample(count).to_list()) == count)
+        
+        linq = linque.Linque(d for d in data)
+        self.assertTrue(len(linq.sample(count).to_list()) == count)
+    
+    
     def test_select(self):
         """Tests whether select works correctly."""
         
@@ -513,6 +640,31 @@ class TestCase(unittest.TestCase):
         self.assertEqual(linq.select(lambda d: d[1]).to_tuple(), (0, 10, 20, 30, 40))
     
     
+    def test_select_many(self):
+        """Tests whether select_many works correctly."""
+        
+        data = ((0, 0), (1, 10), (2, 20), (3, 30), (4, 40))
+        
+        linq = linque.Linque(data)
+        self.assertEqual(linq.select_many(lambda d: d).to_tuple(), (0, 0, 1, 10, 2, 20, 3, 30, 4, 40))
+        
+        linq = linque.Linque(d for d in data)
+        self.assertEqual(linq.select_many(lambda d: d).to_tuple(), (0, 0, 1, 10, 2, 20, 3, 30, 4, 40))
+    
+    
+    def test_shuffle(self):
+        """Tests whether shuffle works correctly."""
+        
+        data = (0, 1, 2, 3, 4)
+        count = len(data)
+        
+        linq = linque.Linque(data)
+        self.assertTrue(len(linq.shuffle().to_list()) == count)
+        
+        linq = linque.Linque(d for d in data)
+        self.assertTrue(len(linq.shuffle().to_list()) == count)
+    
+    
     def test_single(self):
         """Tests whether single works correctly."""
         
@@ -527,10 +679,16 @@ class TestCase(unittest.TestCase):
         data = (0, 1, 2, 3, 4, 5, 6, 7, 8, 9)
         
         linq = linque.Linque(data)
-        self.assertEqual(linq.single(lambda d: 3<d<5), 4)
+        self.assertEqual(linq.single(lambda d: 3 < d < 5), 4)
         
         linq = linque.Linque(d for d in data)
-        self.assertEqual(linq.single(lambda d: 3<d<5), 4)
+        self.assertEqual(linq.single(lambda d: 3 < d < 5), 4)
+        
+        linq = linque.Linque(data)
+        self.assertEqual(linq.single(lambda d: d > 10, -1), -1)
+        
+        linq = linque.Linque(d for d in data)
+        self.assertEqual(linq.single(lambda d: d > 10, -1), -1)
         
         linq = linque.Linque(data)
         with self.assertRaises(ValueError):
@@ -542,11 +700,19 @@ class TestCase(unittest.TestCase):
         
         linq = linque.Linque(data)
         with self.assertRaises(ValueError):
-            linq.single(lambda d: d>5)
+            linq.single(lambda d: d > 5)
         
         linq = linque.Linque(d for d in data)
         with self.assertRaises(ValueError):
-            linq.single(lambda d: d>5)
+            linq.single(lambda d: d > 5)
+        
+        linq = linque.Linque(data)
+        with self.assertRaises(ValueError):
+            linq.single(lambda d: d > 5, -1)
+        
+        linq = linque.Linque(d for d in data)
+        with self.assertRaises(ValueError):
+            linq.single(lambda d: d > 5, -1)
     
     
     def test_skip(self):
@@ -583,44 +749,44 @@ class TestCase(unittest.TestCase):
         
         linq = linque.Linque(d for d in data)
         self.assertEqual(linq.sort().to_tuple(), (0, 1, 2, 3, 4, 5, 6, 7, 8, 9))
-    
-    
-    def test_sort_by(self):
-        """Tests whether sort_by works correctly."""
+        
+        linq = linque.Linque(data)
+        self.assertEqual(linq.sort(reverse=True).to_tuple(), (9, 8, 7, 6, 5, 4, 3, 2, 1, 0))
+        
+        linq = linque.Linque(d for d in data)
+        self.assertEqual(linq.sort(reverse=True).to_tuple(), (9, 8, 7, 6, 5, 4, 3, 2, 1, 0))
         
         data = ((1, 8), (2, 0), (3, 2), (4, 3), (5, 5), (6, 1), (7, 6), (8, 7), (9, 4), (0, 9))
         
         linq = linque.Linque(data)
-        self.assertEqual(
-            linq.sort_by(lambda d: d[1]).to_tuple(), ((2, 0), (6, 1), (3, 2), (4, 3), (9, 4), (5, 5), (7, 6), (8, 7), (1, 8), (0, 9)))
+        self.assertEqual(linq.sort(lambda d: d[1]).to_tuple(), ((2, 0), (6, 1), (3, 2), (4, 3), (9, 4), (5, 5), (7, 6), (8, 7), (1, 8), (0, 9)))
         
         linq = linque.Linque(d for d in data)
-        self.assertEqual(
-            linq.sort_by(lambda d: d[1]).to_tuple(), ((2, 0), (6, 1), (3, 2), (4, 3), (9, 4), (5, 5), (7, 6), (8, 7), (1, 8), (0, 9)))
-    
-    
-    def test_sort_desc(self):
-        """Tests whether sort_desc works correctly."""
-        
-        data = (8, 0, 2, 3, 5, 1, 6, 7, 4, 9)
+        self.assertEqual(linq.sort(lambda d: d[1]).to_tuple(), ((2, 0), (6, 1), (3, 2), (4, 3), (9, 4), (5, 5), (7, 6), (8, 7), (1, 8), (0, 9)))
         
         linq = linque.Linque(data)
-        self.assertEqual(linq.sort_desc().to_tuple(), (9, 8, 7, 6, 5, 4, 3, 2, 1, 0))
+        self.assertEqual(linq.sort(lambda d: d[1], True).to_tuple(), ((0, 9), (1, 8), (8, 7), (7, 6), (5, 5), (9, 4), (4, 3), (3, 2), (6, 1), (2, 0)))
         
         linq = linque.Linque(d for d in data)
-        self.assertEqual(linq.sort_desc().to_tuple(), (9, 8, 7, 6, 5, 4, 3, 2, 1, 0))
-    
-    
-    def test_sort_by_desc(self):
-        """Tests whether sort_by_desc works correctly."""
+        self.assertEqual(linq.sort(lambda d: d[1], True).to_tuple(), ((0, 9), (1, 8), (8, 7), (7, 6), (5, 5), (9, 4), (4, 3), (3, 2), (6, 1), (2, 0)))
         
-        data = ((1, 8), (2, 0), (3, 2), (4, 3), (5, 5), (6, 1), (7, 6), (8, 7), (9, 4), (0, 9))
+        data = ((1, "d", 11), (0, "a", 10), (0, "c", 100), (0, "b", 100), (1, "e", 10), (0, "b", 1000), (2, "f", 20))
+        model = ((2, "f", 20), (1, "d", 11), (1, "e", 10), (0, "a", 10), (0, "b", 1000), (0, "b", 100), (0, "c", 100))
         
         linq = linque.Linque(data)
-        self.assertEqual(linq.sort_by_desc(lambda d: d[1]).to_tuple(), ((0, 9), (1, 8), (8, 7), (7, 6), (5, 5), (9, 4), (4, 3), (3, 2), (6, 1), (2, 0)))
+        self.assertEqual(linq.sort(reverse=[True, False, True]).to_tuple(), model)
         
         linq = linque.Linque(d for d in data)
-        self.assertEqual(linq.sort_by_desc(lambda d: d[1]).to_tuple(), ((0, 9), (1, 8), (8, 7), (7, 6), (5, 5), (9, 4), (4, 3), (3, 2), (6, 1), (2, 0)))
+        self.assertEqual(linq.sort(reverse=[True, False, True]).to_tuple(), model)
+        
+        data = ((1, "d", 11), (0, "a", 10), (0, "c", 100), (0, "b", 100), (1, "e", 10), (0, "b", 1000), (2, "f", 20))
+        model = ((0, "a", 10), (0, "b", 1000), (0, "b", 100), (0, "c", 100), (1, "d", 11), (1, "e", 10), (2, "f", 20))
+        
+        linq = linque.Linque(data)
+        self.assertEqual(linq.sort(lambda d: (d[1], d[2]), reverse=[False, True]).to_tuple(), model)
+        
+        linq = linque.Linque(d for d in data)
+        self.assertEqual(linq.sort(lambda d: (d[1], d[2]), reverse=[False, True]).to_tuple(), model)
     
     
     def test_sum(self):
@@ -750,21 +916,14 @@ class TestCase(unittest.TestCase):
         linq = linque.Linque(d for d in data1)
         items2 = (d for d in data2)
         self.assertEqual(linq.union(items2).to_tuple(), ((0, 1), (0, 2), (1, 1), (1, 2), (0, 3)))
-    
-    
-    def test_union_by(self):
-        """Tests whether union_by works correctly."""
-        
-        data1 = ((0, 1), (0, 1), (0, 2))
-        data2 = ((1, 1), (1, 2), (1, 2), (0, 3))
         
         linq = linque.Linque(data1)
         items2 = data2
-        self.assertEqual(linq.union_by(items2, lambda d: d[1]).to_tuple(), ((0, 1), (0, 2), (0, 3)))
+        self.assertEqual(linq.union(items2, lambda d: d[1]).to_tuple(), ((0, 1), (0, 2), (0, 3)))
         
         linq = linque.Linque(d for d in data1)
         items2 = (d for d in data2)
-        self.assertEqual(linq.union_by(items2, lambda d: d[1]).to_tuple(), ((0, 1), (0, 2), (0, 3)))
+        self.assertEqual(linq.union(items2, lambda d: d[1]).to_tuple(), ((0, 1), (0, 2), (0, 3)))
     
     
     def test_where(self):
